@@ -1,14 +1,14 @@
-from __future__ import annotations
 from lxml import etree
 import logging
 import re
 import inspect
 
-logging.basicConfig(level=logging.INFO)
 
-class NS:
+logging.getLogger(__name__)
+
+class Node:
     '''
-    Class for adding a namespace attribute to other classes
+    Class that generically represents an xml/yang node
     '''
     _ns = None
 
@@ -22,31 +22,40 @@ class NS:
         else:
             return None
 
-class Container(NS):
+
+
+class Container(Node):
 
     '''
     Class meant to represent a Yang container.
-    Only instance initializer classes are allowed to set public attributes
+    Basically this is a 'Node' class that can have children.
+    Only instance initializer classes are allowed to set children
     '''
     def __setattr__(self, name, value):
         '''
-        Limiting who can set attribues and how.
+        Limiting who can set container children and how.
         '''
-
         # private attributes can be set with no restriction,
         # otherwise we'd be stepping on ourselves
         if name.startswith("_"):
             self.__dict__[name] = value
             return
 
+        # if the attribute exists already and we're setting it as the right
+        # type, that is fine too
+        if name in self.__dict__ and type(self.__dict__[name]) == type(value):
+            self.__dict__[name] = value
+            return
+
         # Only the initializer or node functions can create new attribues.
         # looking up who is trying to set an attribute
         caller = inspect.currentframe().f_back
         caller_info = inspect.getframeinfo(caller)
         calling_func = caller_info[2]
 
-        # can only initialize attributes as types Container, LeafList, or Leaf
-        if(calling_func == '__init__'):
+        # only the special ncs_instance function 'initialize_model' can set
+        # children on containers
+        if(calling_func == 'initialize_model'):
 
             # any assigned value must of one of our special 'yang' types
             if isinstance(value, (Container, LeafList, List, Leaf)):
@@ -54,21 +63,30 @@ class Container(NS):
             else:
                 raise TypeError(f'Can only initialize attributes as Leafs or Nodes')
 
-        # any other direct assignment attribute (outside of choose) must
-        # be to a leaf
-        elif(calling_func != 'choose'):
-            if not(isinstance(self.__dict__[name], Leaf)):
-                raise TypeError(f'Can only directly set attributes on Leafs!')
-            self.__dict__[name].value = value
-        else:
+
+        # a user can call the 'choice' function on a choice child
+        elif(calling_func == 'choose'):
             self.__dict__[name] = value
 
+        # the user can also directly assign a value to a leaf child
+        elif(isinstance(self.__dict__[name], Leaf)):
+            self.__dict__[name].value = value
+
+        # or they can directly assign a list to a leaflist
+        elif isinstance(self.__dict__[name], LeafList) and type(value) == list:
+                self.__dict__[name]._values = []
+
+                # assigning each entry in our list individually
+                # so that the type is checked
+                for v in value:
+                    self.__dict__[name].append(v)
 
     def __getattribute__(self, name):
+
         '''
         Overloading 'getattribute' so that if someone wants to read
         a Leaf or LeafList attribute they'll get the value
-        of the Leaf
+        of the Node
         '''
 
         obj = super().__getattribute__(name)
@@ -88,7 +106,8 @@ class Container(NS):
                 children[name] = obj
         return children
 
-class Leaf(NS):
+
+class Leaf(Node):
     '''
     Leaf class that takes a type in its constructor.
     '''
@@ -103,15 +122,14 @@ class Leaf(NS):
     @value.setter
     def value(self, value):
         if type(value) != self._type:
-            raise TypeError('Invalid type for {value}, must be {self._type}')
+            raise TypeError(f'Invalid type for {value}, must be {self._type}')
    
         self._value = value
 
 
-
-class LeafList(NS):
+class LeafList(Node):
     '''
-    Append-only list class that takes a type in its constructor.
+    LeafList class that takes a type in its constructor.
     Newly appended items must be of that type
     '''
     def __init__(self, leaf_type:type):
@@ -128,6 +146,9 @@ class LeafList(NS):
         for v in values:
             self.append(v)
 
+    def remove(self, value):
+        self._values.remove(value)
+
     def __iter__(self):
         return(iter(self._values))
 
@@ -135,25 +156,27 @@ class LeafList(NS):
     def values(self):
         return self._values
 
-class List(NS):
+
+class List(Node):
     '''
     Dict class that takes a class type in its constructor.
     Newly-appended items must be of that class.
     '''
-    def __init__(self, list_type:Container, keyattr:str='name'):
+    def __init__(self, list_type:'Container', keyattr:str='name'):
         self._class = list_type()
+        self._class.initialize_model()
         self._values = {}
         self._keyattr = keyattr
 
         if keyattr not in self._class.__dict__:
-            raise ValueError(f'{keyattr} not an attribute in {class_instance}') 
+            raise ValueError(f'{keyattr} not an attribute in {self._class}') 
 
     def __getitem__(self, key):
         return self._values[key]
 
     def __setitem__(self, key, value):
 
-        # item must be of the write class type
+        # item must be of the right class type
         if(not(isinstance(value, self._class.__class__))):
             raise TypeError(f'Invalid value type, must be {self._class}')
 
@@ -170,6 +193,8 @@ class List(NS):
         return self._values.keys()
     def values(self):
         return self._values.values()
+    def pop(self, key):
+        return self._values.pop(key)
     
     def __iter__(self):
         return(iter(self._values))
@@ -179,17 +204,14 @@ class Choice(Container):
     Yang 'choice' object model. Basically a specialized container where
     only some of the children are valid.
 
-    Practically speaking, you can set attributes for both choices
-    before saying which choice is valid. NCSInstance.gen_xml
+    Practically speaking, you can set attributes for any/all choices, no 
+    matter what you 'choose' w/the choose method. NCSInstance.gen_xml
     does validate that you have chosen which path is valid (by that we mean
     you set the choice attribute) so it knows what to output.
     '''
     def __init__(self, choices:list, choice=None):
         self._choices = [self._pythonify(c) for c in choices]
 
-        for c in self._choices:
-            self.__dict__[c] = Container()
-
         if choice:
             self.choose(choice)
         else:
@@ -201,11 +223,6 @@ class Choice(Container):
         if choice not in self._choices:
             raise ValueError(f'Invalid choice {choice} for {self}')
 
-        for c in self._choices:
-            if c != choice and c in self.__dict__:
-                delattr(self,c)
-
-        self._choices = [choice]
         self._choice = choice
 
     @property
@@ -244,6 +261,10 @@ class NCSInstance(Container):
     _nsmap = {
             "config":"http://tail-f.com/ns/config/1.0",
             }
+    def __init__(self):
+        self.initialize_model()
+
+
 
     def _get_path(self):
         '''
@@ -268,6 +289,11 @@ class NCSInstance(Container):
 
         return(fq_path)
 
+    def __eq__(self, other):
+        if isinstance(other, NCSInstance):
+            return self.gen_xml() == other.gen_xml()
+        return False
+
     def gen_xml(self):
         '''
         Generate xml data based on our NCS instance object.
@@ -301,6 +327,8 @@ class NCSInstance(Container):
         return xml_str
 
     def _walk_tree(self, tree:etree.Element, node_name:str, node_obj):
+
+        logging.debug(f'processing node {node_name}, type {type(node_obj)}')
        
         # un-pythonify node names
         node_name = node_name.replace("_","-")
@@ -311,6 +339,7 @@ class NCSInstance(Container):
 
         # Leaf processing - just need to extract leaf value and set as text
         if isinstance(node_obj, Leaf):
+            logging.debug(f'saving leaf value {node_obj.value}')
             if node_obj.value != None:
                 xml_node = etree.SubElement(tree, node_name, nsmap=node_obj.nsmap)
                 xml_node.text = str(node_obj.value)
@@ -318,6 +347,7 @@ class NCSInstance(Container):
         
         # LeafList - extract leaf values for all items in list
         if isinstance(node_obj, LeafList):
+            logging.debug(f'saving leaflist values {list(node_obj)}')
             for n in node_obj:
                 xml_node = etree.SubElement(tree, node_name, nsmap=node_obj.nsmap)
                 xml_node.text = str(n)
@@ -327,6 +357,7 @@ class NCSInstance(Container):
         if isinstance(node_obj, List):
 
             for entry_name, entry_data in node_obj.items():
+                logging.debug(f'processing list object {entry_name}')
                 xml_list_entry = etree.SubElement(tree, node_name, nsmap=node_obj.nsmap)
                 children = entry_data.get_child_nodes()
                 for name, obj in children.items():
@@ -340,8 +371,10 @@ class NCSInstance(Container):
             if(node_obj.choice==None):
                 raise ValueError(f'Need to set {node_name} choice!')
 
-            choice = node_obj.choice
-            self._walk_tree(tree, node_obj.choice, getattr(node_obj, choice))
+            child_name = node_obj.choice
+            child_obj = node_obj.__dict__[child_name]
+            logging.debug(f'choice is {child_name} {type(child_obj)}')
+            self._walk_tree(tree, child_name, child_obj)
 
         # container is a container
         elif isinstance(node_obj, Container):
